"""Pytest configuration"""
import pytest  # type: ignore

from ridesystems.reports import Reports
from ridesystems.api import API


def pytest_addoption(parser):
    """Adds command line options"""
    parser.addoption("--username", action="store")
    parser.addoption("--password", action="store")


@pytest.fixture(name='username')
def username_fixture(request):
    """Username fixture"""
    return request.config.getoption("username")


@pytest.fixture(name='password')
def password_fixture(request):
    """Password fixture"""
    return request.config.getoption("password")


@pytest.fixture(name='reports_fixture')
def fix_report(username, password):
    """Setup for the tests"""
    assert username, "Username required (use --username)"
    assert password, "Password required (use --password)"
    return Reports(username, password)


@pytest.fixture
def ridesystems_api():
    """Fixture for the Ridesystems API object"""
    return API('xx')  # Ridesystems doesn't actually check the API key



""""
The Ridesystems API, to access workable data of any Ridesystems system
Written by: Ritwik Gupta (2015)
Updated by: Brian Seel (2020) brian.seel@baltimorecity.gov

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
# pylint:disable=too-few-public-methods,inherit-non-class
from typing import Dict, Any, Optional, TypedDict, Union, List, cast
from datetime import datetime, date

from tenacity import retry, wait_random_exponential, stop_after_attempt
import requests


# Return types used for type checking
class StopArrivalTimesDict(TypedDict):
    """Used for type checking"""
    timesPerStop: int
    routeIDs: Optional[str]
    routeStopIDs: Optional[str]
    ApiKey: Optional[str]


class VehicleRouteStopEstimates(TypedDict):
    """Used for type checking"""
    quantity: str
    vehicleIdStrings: Optional[str]
    ApiKey: Optional[str]


class RouteSchedules(TypedDict):
    """Used for type checking"""
    routeID: Optional[int]
    ApiKey: Optional[str]


RidershipDate = Union[date, datetime]


class Ridership(TypedDict):
    """Used for type checking"""
    StartDate: RidershipDate
    EndDate: RidershipDate
    ApiKey: Optional[str]


ApiDataTypes = Union[VehicleRouteStopEstimates, StopArrivalTimesDict, RouteSchedules, Ridership]
QueryResult = Dict[str, Any]
QueryResultList = List[QueryResult]


class API:
    """Wrapper for the Ride Systems API. Their API is pretty terrible, so a lot of functionality is in the scraper from
    ridesystems.reports.Reports"""

    def __init__(self, api_key: str, base_url="https://cityofbaltimore.ridesystems.net"):
        """
        :param api_key: Ridesystems API key (although nothing really happens if you don't provide one. RS doesn't check)
        :param base_url: Base url of your Ride Systems instance
        """
        self.api_key = api_key
        self.base_url = base_url
        self.session = requests.session()

    @retry(wait=wait_random_exponential(multiplier=1, max=60), stop=stop_after_attempt(7), reraise=True)
    def get_routes_for_map_with_schedule_with_encoded_line(self) -> QueryResultList:
        """
        Used to retrieve active Routes from Ride Systems. Also contains a link to a schedule for each particular route

        :return RouteWithSchedule: List of dictionaries with the following information
            RouteID – Unique identifier for the route
            Description – Description of the route
            TextingKey – Unique string that represents the route when texted to Ride Systems
            MapLineColor – Color of the route on the map
            MapZoom – Default zoom level to set when this route is selected
            MapLatitude – Default latitude to set when this route is selected
            MapLongitude – Default longitude to set when this route is selected
            IsVisibleOnMap – Reflects whether this route should be shown on the main page
            IsCheckedOnMap –Reflects whether this route should be checked by default on the main page
            IsCheckLineOnlyOnMap– Does not show on app
            Order – Where this route falls in order with the other routes
            StopTimesPDFLink – Link to use to get the schedule for this route
            HideRouteLine – Setting on whether to show the route line
            ShowPolygon– Setting to show the route line as a polygon rather than a line
            InfoText– Subline text to use in the map panel for this route
            EncodedPolyline – Google encoded points that make up the polyline
            ETATypeID – What times do we show for the stops: 1) Estimates, 2) Schedules, or 3) SchedulesWithEstimates?
                        1. Estimates – Show the Estimate time (in time of day or minutes to arrival based on
                            EstimateDisplayType)
                        2. Schedules – Show the Scheduled time (in time of day or minutes to arrival based on
                            EstimateDisplayType)
                        3. Schedule With Estimates – Show the later of either the Schedule Time or Estimate time (in
                            time of day or minutes to arrival based on EstimateDisplayType)
            ShowRouteArrows– Setting on whether to show arrows on the Route Line
            UseScheduleTripsInPassengerCounter– Not used
            VehicleMarkerCssClass– CSS Class for Vehicle Marker
            Stops - Name of the stop
            RouteStopID– Unique Identifier for a Stop on a Route
            RouteID – Unique Identifier for a Route
            RouteDescription – Description of the Route
            Description – Description of the Stop
            Order – Order of this stop in comparison with the other stops.
            Heading – GPS Heading taken when the bus moves away from this stop.
            SecondsAtStop – Number of seconds the bus stays at this stop before moving to the next
            SecondsToNextStop – Number of seconds it takes the bus from move from this stop to the next.
            ShowEstimatesOnMap – Reflects setting on whether to show the estimates for this stop on the mapcpanel
            ShowDefaultedOnMap –Reflects whether to show this stop on the map
            TextingKey – Unique string that represents the Stop when texted to Ride Systems
            MaxZoomLevel –Reflects the zoom level to start showing this stop to the user
            SignVerbiage– Alternate Stop Description used by Sign Applications
            MapPoints– Not used
            Landmarks – Misc landmarks associated with this particular route
            LandmarkID – Unique identifier for the Landmark
            Label – Description to be put on the map
            AddressID – Unique ID associated with the address of this Landmark
            Latitude
            Longitude
        """
        response = self.session.get(
            "{}/Services/JSONPRelay.svc/GetRoutesForMapWithScheduleWithEncodedLine".format(self.base_url))
        return response.json()

    @retry(wait=wait_random_exponential(multiplier=1, max=60), stop=stop_after_attempt(7), reraise=True)
    def get_map_vehicle_points(self) -> QueryResultList:
        """
        Return the map location for all active vehicles.

        :return vehicle: (list) List of dictionaries containing the following information
            VehicleID — Unique Identifier for a Vehicle
            RouteID – Unique Identifier for a Route
            Name – Name of the Vehicle
            Latitude – Latitude of Vehicle’s current position
            Longitude – Longitude of Vehicle’s current position
            GroundSpeed – Speed of Vehicle
            Heading – Heading of Vehicle
            Seconds – Seconds since the vehicle reported its location
            IsOnRoute – Is the vehicle on Route?
            IsDelayed– Is the vehicle Delayed?
        """
        response = self.session.get("{}/Services/JSONPRelay.svc/GetMapVehiclePoints".format(self.base_url))
        return response.json()

    @retry(wait=wait_random_exponential(multiplier=1, max=60), stop=stop_after_attempt(7), reraise=True)
    def get_vehicle_route_stop_estimates(self, vehicle_id: List[int],
                                         quantity: int = 2) -> QueryResultList:
        """Return {quantity} stop estimates for all active vehicles.

        :param vehicle_id: (list) List of integers of Vehicle ID’s to retrieve
        :param quantity: (int) Number of records to return.

        :return vehicleestimates: The following information
            VehicleID — Unique Identifier for a Vehicle
            Estimates[] – Estimates
            RouteStopId – Unique ID of each Route Stop
            Description – Description of Route Stop
            OnRoute – Is the vehicle on Route?
            VehicleId – ID of Vehicle
            Text – Text of Estimate
            Time –Time of Day of expected arrival
            Seconds – Estimated Seconds until Arrival
            IsArriving – Is the vehicle arriving?
            EstimateTime – Estimated arrival Time of Day
            ScheduledTime – Scheduled arrival Time of Day
            OnTimeStatus– 0 – On time, 2 – Early, 3 – Late

        """
        payload: VehicleRouteStopEstimates = {"quantity": str(quantity),
                                              "vehicleIdStrings": None,
                                              "ApiKey": self.api_key}

        if vehicle_id:
            payload["vehicleIdStrings"] = ",".join(str(i) for i in vehicle_id)
        response = self.session.get("{}/Services/JSONPRelay.svc/GetVehicleRouteStopEstimates".format(self.base_url),
                                    params=cast(Dict, payload))
        return response.json()

    @retry(wait=wait_random_exponential(multiplier=1, max=60), stop=stop_after_attempt(7), reraise=True)
    def get_stop_arrival_times(self, times_per_stop: int = 1,
                               route_ids: List[int] = None,
                               stop_ids: List[int] = None) -> QueryResultList:
        """
        Return stop arrival times for all vehicles.

        :param times_per_stop: (int) Optional, number of scheduled times to return.
        :param route_ids: (list) Optional, to restrict the results to given Route ID(s).
        :param stop_ids: (list of ints) Optional, to restrict the results to given Stop ID(s)

        :return RouteStopArrival: (List of dictionaries) the following information
            RouteID – Unique Identifier for a Route
            RouteStopID – Unique identifier for a Route Stop
            Description – Not used
            Times[] – Arrival Times
            VehicleId – ID of Vehicle
            Text – Text of Estimate
            Time – Time of Day of expected arrival
            Seconds – Estimated Seconds till Arrival
            IsArriving – Is the vehicle arriving?
            EstimateTime – Estimated arrival Time of Day
            ScheduledTime – Scheduled arrival or departure Time of Day
            IsDeparted – Has the vehicle departed?
            ScheduledArrivalTime – Scheduled arrival Time of Day
            ScheduledDepartureTime – Scheduled departure Time of Day
            OnTimeStatus– 0 – On time, 2 – Early, 3 – Late
        """
        payload: StopArrivalTimesDict = {"timesPerStop": times_per_stop,
                                         "routeIDs": None,
                                         "routeStopIDs": None,
                                         "ApiKey": self.api_key}
        if route_ids:
            payload["routeIDs"] = ", ".join([str(x) for x in route_ids])

        if stop_ids:
            payload["routeStopIDs"] = ", ".join(map(str, stop_ids))

        response = self.session.get("{}/Services/JSONPRelay.svc/GetStopArrivalTimes".format(self.base_url),
                                    params=cast(Dict, payload))
        return response.json()

    @retry(wait=wait_random_exponential(multiplier=1, max=60), stop=stop_after_attempt(7), reraise=True)
    def get_route_stop_arrivals(self, times_per_stop: int = 1) -> QueryResultList:
        """
        Return stop arrival times for all vehicles.
        :param times_per_stop: (int) Optional, number of scheduled times to return.

        :return:
            RouteID - value representing the route
            RouteStopID - value representing the stop
            ScheduledTimes - dictionary with the following values for the RouteStopID
                ArrivalTimeUTC - scheduled arrival tim->e
                AssignedVehicleId - VehicleID assigned to the block with this scheduled arrival
                Block - string identifying the block assigned to the next arrival
                DepartureTimeUTC - scheduled stop departure time
            VehicleEstimates - dictionary with the following values for each VehicleID
                Block - string identifying the schedule block
                OnRoute - is the vehicle on the assigned route
                SecondsToStop - seconds until we get to RouteStopID
                VehicleId – ID of Vehicle
        """

        payload: StopArrivalTimesDict = {"timesPerStop": times_per_stop,
                                         "routeIDs": None,
                                         "routeStopIDs": None,
                                         "ApiKey": self.api_key}
        response = self.session.get("{}/Services/JSONPRelay.svc/GetRouteStopArrivals".format(self.base_url),
                                    params=cast(Dict, payload))
        return response.json()

    @retry(wait=wait_random_exponential(multiplier=1, max=60), stop=stop_after_attempt(7), reraise=True)
    def get_route_schedules(self, route_id: int = None) -> QueryResultList:
        """
        Used to return scheduled times for a Route. This is used for cyclical routes, where the route runs twice an
        hour on the exact same path and schedule.

        :param route_id: (int) Optional, to restrict the results to a given Route ID

        :return RouteSchedule: (list of dictionaries)
            RouteScheduleID – Unique ID for each schedule
            RouteID – Unique Identifier for a Route
            StartTime – Time of day the Route Schedule starts
            StartTimeUTC – UTC Time of day the Route Schedule starts
            EndTime – Time of day the Route Schedule ends
            EndTimeUTC – UTC Time of day the Route Schedule ends
            LoopsPerHour – Number of loops run per hour
            StopTimes -
            RouteStopScheduleID – Unique ID for each stop on the schedule
            RouteScheduleID – Unique ID for the schedule
            RouteStopID – Unique Identifier for a Stop on a Route
            MinutesAfterStart – Number of minutes after the start of the loop until arrival at this stop
        """
        payload: RouteSchedules = {"routeID": route_id, "ApiKey": self.api_key}
        response = self.session.get("{}/Services/JSONPRelay.svc/GetRouteSchedules".format(self.base_url),
                                    params=cast(Dict, payload))
        return response.json()

    @retry(wait=wait_random_exponential(multiplier=1, max=60), stop=stop_after_attempt(7), reraise=True)
    def get_route_schedule_times(self, route_id: int = None) -> QueryResultList:
        """
        Used to return times in the day that a Route is active.

        :param route_id: (int) Optional, to restrict the results to a given Route ID.

        :return RouteScheduleTime: (list of dictionaries)
            RouteID – Unique Identifier for a Route
            StartTimeUTC – UTC time of day the Route Schedule starts
            EndTimeUTC – UTC time of day the Route Schedule ends
            StartTime – Time of day (in MST) the Route Schedule starts
            EndTime – Time of day (in MST) the Route Schedule ends
            ServerTime– Current time of day (in MST)
            ServerTimeUTC– UTC Current Time of day
        """
        payload: RouteSchedules = {"routeID": route_id, "ApiKey": self.api_key}
        response = self.session.get("{}/Services/JSONPRelay.svc/GetRouteScheduleTimes".format(self.base_url),
                                    params=cast(Dict, payload))
        return response.json()

    @retry(wait=wait_random_exponential(multiplier=1, max=60), stop=stop_after_attempt(7), reraise=True)
    def get_routes(self, route_id: int = None) -> QueryResultList:
        """
        Abbreviated view of all active Routes on Ride Systems. Used for Smart Phones where data size is a limiting
        factor.

        :param route_id: (int) Optional, to restrict the results to a given Route ID.

        :return SmartPhoneRoute: (list of dictionaries)
            RouteID – Unique Identifier for a Route
            Description – Description of the Route
            MapLineColor – Color of the Route on the Map
            MapZoom – Default Zoom level to set when this route is selected
            MapLatitude – Default Latitude to set when this route is selected
            MapLongitude – Default Longitude to set when this route is selected
            IsVisibleOnMap – Reflects if this route should be shown on the main page
            StopTimesPDFLink – Link to use to get the Schedule for this Route
            HideRouteLine – Setting on whether to show the Route Line
            UseScheduleTripsInPassengerCounter– Not used
        """
        payload: RouteSchedules = {"routeID": route_id, "ApiKey": self.api_key}
        response = self.session.get("{}/Services/JSONPRelay.svc/GetRoutes".format(self.base_url),
                                    params=cast(Dict, payload))
        return response.json()

    @retry(wait=wait_random_exponential(multiplier=1, max=60), stop=stop_after_attempt(7), reraise=True)
    def get_stops(self, route_id: int = None) -> QueryResultList:
        """
        Abbreviated view of all active Stops on a route. Used for Smart Phones where data size is a limiting factor.

        :param route_id: (int) Optional, to restrict the results to a given Route ID.

        :return SmartPhoneRouteStop: (list of dictionaries)
            RouteStopID – Unique Identifier for a Stop on a Route
            RouteID – Unique Identifier for a Route
            Description – Description of the Stop
            Longitude
            Latitude
            TextingKey – Unique string that represents the Stop when Texted to Ride Systems
            MaxZoomLevel –Reflects the zoom level to start showing this stop to the user
            ShowEstimatesOnMap – Reflects setting on whether to show the estimates for this stop on the map
            panel
            ShowDefaultedOnMap – Reflects whether to show this stop on the map
            MapPoints – GPS points that make up the path to the next stop
            Latitude
            Longitude
            Heading- Not used
        """
        payload: RouteSchedules = {"routeID": route_id, "ApiKey": self.api_key}
        response = self.session.get("{}/Services/JSONPRelay.svc/GetStops".format(self.base_url),
                                    params=cast(Dict, payload))
        return response.json()

    @retry(wait=wait_random_exponential(multiplier=1, max=60), stop=stop_after_attempt(7), reraise=True)
    def get_markers(self, route_id: int = None) -> QueryResultList:
        """
        Abbreviated view of all active Landmarks on a route. Used for Smart Phones where data size is a limiting factor.

        :param route_id: (int) Optional, to restrict the results to a given Route ID.

        :return SmartPhoneLandmark: (list of dictionaries)
            RouteID – Unique Identifier for a Route
            LandmarkID – Unique identifier for the Landmark
            Label – Description to be put on the map
            Latitude
            Longitude
        """
        payload: RouteSchedules = {"routeID": route_id, "ApiKey": self.api_key}
        response = self.session.get("{}/Services/JSONPRelay.svc/GetMarkers".format(self.base_url),
                                    params=cast(Dict, payload))
        return response.json()

    @retry(wait=wait_random_exponential(multiplier=1, max=60), stop=stop_after_attempt(7), reraise=True)
    def get_map_config(self) -> QueryResult:
        """
        Returns settings that are used for laying out the map

        :return MapConfig: See docs for return information
        """
        response = self.session.get("{}/Services/JSONPRelay.svc/GetMapConfig".format(self.base_url))
        return response.json()

    @retry(wait=wait_random_exponential(multiplier=1, max=60), stop=stop_after_attempt(7), reraise=True)
    def get_routes_for_map(self) -> QueryResultList:
        """Return the routes with Vehicle Route Name, Vehicle ID, and all stops, etc."""
        response = self.session.get("{}/Services/JSONPRelay.svc/GetRoutesForMap".format(self.base_url))
        return response.json()

    @retry(wait=wait_random_exponential(multiplier=1, max=60), stop=stop_after_attempt(7), reraise=True)
    def get_ridership_data(self, start_date: RidershipDate,
                           end_date: RidershipDate) -> QueryResultList:
        """
        Return the ridership from an APC

        :param start_date: The date to begin the range (either in the format 2020-03-01 or 2020-03-01 1:00 PM)
        :type start_date: str
        :param end_date: The date to end the range (same format as start_date)
        :type end_date: str

        :return Ridership: (list of dictionaries)
            ClientTime
            Counter
            CounterType
            Entries
            Exits
            Latitude
            LongBadgeNumber
            Longitude
            Route
            RouteID
            RouteStop
            RouteStoopID
            ShortBadgeNumberUTC
            Time
            Vehicle
            VehicleID
        """
        payload: Ridership = {"StartDate": start_date, "EndDate": end_date, "ApiKey": self.api_key}
        response = self.session.get("{}/Services/JSONPRelay.svc/GetRidershipData".format(self.base_url),
                                    params=cast(Dict, payload))
        return response.json()



"""ridesystems module"""
from importlib.metadata import PackageNotFoundError, version  # pragma: no cover


try:
    # Change here if project is renamed and does not equal the package name
    DIST_NAME = __name__
    __version__ = version(DIST_NAME)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError



"""
Tests reports.py

Should be called as tox -- --username <username> --password <password>
"""

from datetime import date, datetime, time, timedelta

import pandas as pd
import pytest  # type: ignore

from ridesystems.reports import Reports


def test_login_failure():
    """Tests for login with bad username/password, which throws an exception"""
    with pytest.raises(AssertionError):
        Reports('invalidusername', 'invalidpassword')


def test_get_otp_all(reports_fixture):
    """
    Validate that we get valid data when we pull the on time performance data

    Expected format
    [[date, route, block_id, vehicle, stop, scheduled_dept_time, actual_dept_time],
    ...]
    """
    start_date = datetime.today() - timedelta(days=1)
    end_date = datetime.today() - timedelta(days=1)

    otp_data = reports_fixture.get_otp(start_date, end_date, '11, 12')

    iters = 0
    routes = set()
    stops = set()
    blockid = set()
    ontimestatuses = set()
    vehicles = set()

    for _, row in otp_data.iterrows():
        assert isinstance(row['date'], datetime)
        assert isinstance(row['route'], str)
        assert row['route'] in ['Green', 'Banner', 'Purple', 'Orange']
        assert isinstance(row['stop'], str)
        assert isinstance(row['blockid'], str)
        assert isinstance(row['scheduledarrivaltime'], time) or row['scheduledarrivaltime'] is pd.NaT
        assert isinstance(row['actualarrivaltime'], time) or row['actualarrivaltime'] is pd.NaT
        assert isinstance(row['scheduleddeparturetime'], time) or row['scheduleddeparturetime'] is pd.NaT
        assert isinstance(row['actualdeparturetime'], time) or row['actualdeparturetime'] is pd.NaT
        assert isinstance(row['ontimestatus'], str)
        assert isinstance(row['vehicle'], str)

        routes.add(row['route'])
        stops.add(row['stop'])
        blockid.add(row['blockid'])
        ontimestatuses.add(row['ontimestatus'])
        vehicles.add(row['vehicle'])
        iters += 1

    assert iters > 700
    assert len(routes) >= 4
    assert len(stops) >= 50
    assert len(blockid) >= 4
    assert len(ontimestatuses) == 4
    assert len(vehicles) > 5


def test_get_otp_apr_6(reports_fixture):
    """
    Validate that we get valid data when we pull the on time performance data. April 6 was when RideSystems was doing
    some odd stuff with their data, so its a good edge-case

    Expected format
    [[date, route, block_id, vehicle, stop, scheduled_dept_time, actual_dept_time],
    ...]
    """
    start_date = datetime(2020, 4, 6)
    end_date = datetime(2020, 4, 6)

    otp_data = reports_fixture.get_otp(start_date, end_date, '11, 12')

    iters = 0
    for _, row in otp_data.iterrows():
        assert isinstance(row['date'], datetime)
        assert isinstance(row['route'], str)
        assert row['route'] in ['Green', 'Banner', 'Purple', 'Orange']
        assert isinstance(row['stop'], str)
        assert isinstance(row['blockid'], str)
        assert isinstance(row['scheduledarrivaltime'], time) or row['scheduledarrivaltime'] is pd.NaT
        assert isinstance(row['actualarrivaltime'], time) or row['actualarrivaltime'] is pd.NaT
        assert isinstance(row['scheduleddeparturetime'], time) or row['scheduleddeparturetime'] is pd.NaT
        assert isinstance(row['actualdeparturetime'], time) or row['actualdeparturetime'] is pd.NaT
        assert isinstance(row['ontimestatus'], str)
        assert isinstance(row['vehicle'], str)
        iters += 1

    assert iters > 40


def test_get_runtimes(reports_fixture):
    """Test get_runtimes"""
    start_date = datetime.today() - timedelta(days=1)
    end_date = datetime.today() - timedelta(days=1)

    runtime_data = reports_fixture.get_runtimes(start_date, end_date)

    iters = 0
    for _, row in runtime_data.iterrows():
        assert isinstance(row['start_time'], datetime)
        assert isinstance(row['end_time'], datetime)
        assert isinstance(row['route'], str)
        assert row['route'] in ['Green', 'Banner', 'Purple', 'Orange']
        assert isinstance(row['vehicle'], str)
        iters += 1

    assert iters > 40


def test_get_ridership(reports_fixture):
    """Test get_ridership"""
    ridership_data = reports_fixture.get_ridership(date(2021, 7, 19), date(2021, 7, 19))

    iters = 0
    for _, row in ridership_data.iterrows():
        assert isinstance(row['vehicle'], str)
        assert isinstance(row['route'], str)
        assert row['route'] in ['Green', 'Banner', 'Purple', 'Orange']
        assert isinstance(row['stop'], str)
        assert isinstance(row['latitude'], float)
        assert isinstance(row['longitude'], float)
        assert isinstance(row['datetime'], datetime)
        assert isinstance(row['entries'], int)
        assert isinstance(row['exits'], int)
        iters += 1

    assert iters > 3000


def test_parse_ltiv_data(reports_fixture):
    """Tests the parse_ltiv_data method"""
    expect = {'Return': ('this', 'NONE')}
    actual = reports_fixture.parse_ltiv_data('4|NONE|Return|this|')
    assert compare_ltiv_data(expect, actual), \
        'Did not pass basic assertion. Expected {}. Got {}'.format(expect, actual)

    with pytest.raises(AssertionError):
        reports_fixture.parse_ltiv_data('4|NONE|Return|this')

    expect = {'Return': ('this', 'NONE'), 'Something': ('abcdefghijklmnop', 'NONE')}
    actual = reports_fixture.parse_ltiv_data('4|NONE|Return|this|16|NONE|Something|abcdefghijklmnop|')
    assert compare_ltiv_data(expect, actual), \
        'Did not pass second assertion. Expected {}. Got {}'.format(expect, actual)

    with pytest.raises(AssertionError):
        reports_fixture.parse_ltiv_data('3|NONE|Return|this|7|NONE|x|xxxxxxx|')

    expect = {'Return': ('this', 'NONE'), 'x': ('xxx|xxx', 'NONE')}
    actual = reports_fixture.parse_ltiv_data('4|NONE|Return|this|7|NONE|x|xxx|xxx|')
    assert compare_ltiv_data(expect, actual), \
        'Did not pass extra delimiter test. Expected {}. Got {}'.format(expect, actual)

    expect = {}
    actual = reports_fixture.parse_ltiv_data('')
    assert compare_ltiv_data(expect, actual), 'Empty case failed. Expected {}. Got {}'.format(expect, actual)

    expect = {'': ('', '')}
    actual = reports_fixture.parse_ltiv_data('0||||')
    assert compare_ltiv_data(expect, actual), 'Did not pass basic assertion. Expected {}. Got {}'.format(expect, actual)


def compare_ltiv_data(expected, actual):
    """
    Helper to test the LENGTH|TYPE|ID|VALUE data. It is packed in a dictionary like
    {ID: (VALUE, TYPE)
    """
    for k, val in expected.items():
        actual_v = actual.pop(k)
        if not (actual_v[0] == val[0] and actual_v[1] == val[1]):
            return False
    return actual == {}



"""Test suite for src.api"""
from datetime import date, datetime, timedelta
import logging
import random


def get_random_route(rs_api) -> int:
    """Gets a random route id used for testing"""
    routes = rs_api.get_routes()
    index = random.randint(1, len(routes) - 1)
    route_id = routes[index]['RouteID']
    logging.getLogger().info("Selecting RouteID %s", route_id)
    return route_id


def test_api_get_routes_for_map_with_schedule_with_encoded_line(ridesystems_api):
    """Test for Ridesystems API method get_routes_for_map_with_schedule_with_encoded_line"""
    ret = ridesystems_api.get_routes_for_map_with_schedule_with_encoded_line()
    assert isinstance(ret, list)
    assert ret


def test_api_get_map_vehicle_points(ridesystems_api):
    """Test for Ridesystems API method get_map_vehicle_points"""
    ret = ridesystems_api.get_map_vehicle_points()
    assert isinstance(ret, list)
    assert ret


def test_api_get_vehicle_route_stop_estimates(ridesystems_api):
    """Test for Ridesystems API method get_vehicle_route_stop_estimates"""
    ret = ridesystems_api.get_vehicle_route_stop_estimates([get_random_route(ridesystems_api)])
    assert isinstance(ret, list)
    assert ret


def test_api_get_stop_arrival_times(ridesystems_api):
    """Test for Ridesystems API method get_stop_arrival_times"""
    ret = ridesystems_api.get_stop_arrival_times()
    assert isinstance(ret, list)
    assert ret


def test_api_get_route_stop_arrivals(ridesystems_api):
    """Test for Ridesystems API method get_route_stop_arrivals"""
    ret = ridesystems_api.get_route_stop_arrivals()
    assert isinstance(ret, list)
    assert ret


def test_api_get_route_schedules(ridesystems_api):
    """
    Test for Ridesystems API method get_route_schedules. Ride systems doesn't support this, so we don't check for data.
    """
    ret = ridesystems_api.get_route_schedules([get_random_route(ridesystems_api)])
    assert isinstance(ret, list)
    assert len(ret) >= 0  #


def test_api_get_route_schedule_times(ridesystems_api):
    """
    Test for Ridesystems API method get_route_schedule_times. Ride systems doesn't support this, so we don't check for
    data
    """
    ret = ridesystems_api.get_route_schedule_times()
    assert isinstance(ret, list)
    assert len(ret) >= 0  # Ride systems doesn't support this


def test_api_get_routes(ridesystems_api):
    """Test for Ridesystems API method get_routes"""
    ret = ridesystems_api.get_routes()
    assert isinstance(ret, list)
    assert ret


def test_api_get_stops(ridesystems_api):
    """Test for Ridesystems API method get_stops"""
    ret = ridesystems_api.get_stops()
    assert isinstance(ret, list)
    assert ret


def test_api_get_markers(ridesystems_api):
    """Test for Ridesystems API method get_markers. Ride systems doesn't support this, so we don't check for data."""
    ret = ridesystems_api.get_markers([get_random_route(ridesystems_api)])
    assert isinstance(ret, list)
    assert len(ret) >= 0  # Ridesystems doesn't provide this


def test_api_get_map_config(ridesystems_api):
    """Test for Ridesystems API method get_map_config"""
    ret = ridesystems_api.get_map_config()
    assert isinstance(ret, dict)
    assert ret


def test_api_get_routes_for_map(ridesystems_api):
    """Test for Ridesystems API method get_routes_for_map"""
    ret = ridesystems_api.get_routes_for_map()
    assert isinstance(ret, list)
    assert ret


def test_api_get_ridership_data(ridesystems_api):
    """
    Test for Ridesystems API method get_ridership_data. Ridesystems doesn't have ridership data, so we don't check for
    data
    """
    ret = ridesystems_api.get_ridership_data(date.today(), date.today())
    assert isinstance(ret, list)

    yesterday = datetime.now() - timedelta(days=1)
    ret = ridesystems_api.get_ridership_data(datetime.now(), yesterday)
    assert isinstance(ret, list)



"""
Reports with Ridesystems website

Written independently by brian.seel@baltimorecity.gov. If this were written in collaboration with Ridesystems, there
would just be a reasonable API offered.

CREATE TABLE [ccc_arrival_times2] (
[date] [date] NOT NULL,
[route] varchar(50) NOT NULL,
[stop] varchar(max) NOT NULL,
[blockid] varchar(100) NOT NULL,
[scheduledarrivaltime] [time] NOT NULL,
[actualarrivaltime] [time],
[scheduleddeparturetime] [time] NOT NULL,
[actualdeparturetime] [time],
[ontimestatus] varchar(20),
[vehicle] varchar(50) );
"""
import re
from datetime import date
from io import StringIO
from typing import Any, Dict, List, Tuple, Union

import mechanize  # type: ignore
import pandas as pd  # type: ignore
import requests
from bs4 import BeautifulSoup  # type: ignore
from loguru import logger
from tenacity import retry, wait_random_exponential, stop_after_attempt

HEADERS = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 ' \
          '(KHTML, like Gecko) Chrome/86.0.4240.183 Safari/537.36'


class Reports:
    """Setup for Ridesystems session"""

    def __init__(self, username: str, password: str, baseurl: str = "https://cityofbaltimore.ridesystems.net"):
        self.browser = mechanize.Browser()
        self.browser.addheaders = [('User-agent', HEADERS)]

        self.baseurl = baseurl
        self._login(username, password)

    @retry(wait=wait_random_exponential(multiplier=1, max=60), stop=stop_after_attempt(7), reraise=True)
    def _login(self, username: str, password: str) -> None:
        self.browser.open("{}/login.aspx".format(self.baseurl))
        self.browser.select_form('aspnetForm')

        username_control = self.browser.form.find_control(type='text')
        username_control.value = username
        password_control = self.browser.form.find_control(type='password')
        password_control.value = password

        self.browser.submit()

        # Login validation
        page_contents = self.browser.response().read()
        soup = BeautifulSoup(page_contents, features="html.parser")
        assert soup.find('div', {'class': 'login-panel'}) is None, "Login failed"

    @retry(wait=wait_random_exponential(multiplier=1, max=60), stop=stop_after_attempt(7), reraise=True)
    def _make_response_and_submit(self, ctrl_dict: Dict[str, Union[str, List]], html: str) -> bytes:
        """
        Helper to regenerate a response, assign it to the form, and resubmit it. Used for postbacks
        :param ctrl_dict: Dictionary of page control ids and the values they should be set to
        :return:
        """
        response = mechanize.make_response(html, [('Content-Type', 'text/html')],
                                           self.browser.geturl(), 200, 'OK')
        self.browser.set_response(response)
        self.browser.select_form('aspnetForm')
        self.browser.form.set_all_readonly(False)

        self._set_controls(ctrl_dict)
        return self.browser.submit().read()

    @retry(wait=wait_random_exponential(multiplier=1, max=60), stop=stop_after_attempt(7), reraise=True)
    def get_otp(self, start_date: date, end_date: date,  # pylint:disable=too-many-locals
                hours: str = ','.join([str(x) for x in range(24)])) -> pd.DataFrame:
        """ Pulls the on time performance data
        :param start_date: The start date to search, inclusive. Searches starting from 12:00 AM
        :param end_date: The end date to search, inclusive. Searches ending at 11:59:59 PM
        :param hours: A comma separated list of the hours of the day to search using military time IE '11, 12, 13'
        :return: Returns a dataframe with the keys 'date', 'route', 'stop', 'blockid', 'scheduledarrivaltime',
            'actualarrivaltime', 'scheduleddeparturetime', 'actualdeparturetime', 'ontimestatus', 'vehicle
        """
        logger.info("Getting OTP report for {} to {}", start_date, end_date)
        # Pull the page the first time to get the form that we will need to resubmit a few times
        soup, html = self._select_form(
            '/Secure/Admin/Reports/ReportViewer.aspx?Path='
            '%2fOldRidesystems%2fPerformance+Reports%2fOn+Time+Performance')

        ctrl_dict: Dict[str, Union[str, List]] = {
            # Start Date
            'ctl00$MainContent$ssrsReportViewer$ctl08$ctl03$txtValue': start_date.strftime('%#m/%#d/%Y'),
            # End Date
            'ctl00$MainContent$ssrsReportViewer$ctl08$ctl05$txtValue': end_date.strftime('%#m/%#d/%Y 11:59:59 PM'),
            # Seconds For Early
            'ctl00$MainContent$ssrsReportViewer$ctl08$ctl11$txtValue': '30',
            # Seconds For Late
            'ctl00$MainContent$ssrsReportViewer$ctl08$ctl13$txtValue': '300',
            # Status based on (departure)
            'ctl00$MainContent$ssrsReportViewer$ctl08$ctl15$ddValue': ['1'],
            # Force assign block (No)
            'ctl00$MainContent$ssrsReportViewer$ctl08$ctl17$ddValue': ['2'],
            # Status
            'ctl00$MainContent$ssrsReportViewer$ctl08$ctl19$txtValue': 'On Time,Early,Late,Missing',
            'ctl00$MainContent$ssrsReportViewer$ctl08$ctl19$divDropDown$ctl01$HiddenIndices': '0,1,2,3',
            # Hours
            'ctl00$MainContent$ssrsReportViewer$ctl08$ctl21$txtValue': ','.join([str(x) for x in range(24)]),
            'ctl00$MainContent$ssrsReportViewer$ctl08$ctl21$divDropDown$ctl01$HiddenIndices': hours,
            # Group Data
            'ctl00$MainContent$ssrsReportViewer$ctl08$ctl23$ddValue': ['1'],

            # Other values
            'ctl00$MainContent$ssrsReportViewer$ctl15': 'standards',
            'ctl00$MainContent$ssrsReportViewer$AsyncWait$HiddenCancelField': 'False',
            '__EVENTTARGET': 'ctl00$MainContent$ssrsReportViewer$ctl08$ctl05',
            '__ASYNCPOST': 'true',
            'ctl00$MainContent$scriptManager':
                'ctl00$MainContent$scriptManager|ctl00$MainContent$ssrsReportViewer$ctl08$ctl05'
        }

        self._set_controls(ctrl_dict)
        resp = self.browser.submit().read()
        soup = BeautifulSoup(resp, features='html.parser')

        # set the controls
        routes = [x.text.replace('\xa0', ' ') for x in soup.find_all('label', {
            'for': re.compile(r'ctl00_MainContent_ssrsReportViewer_ctl08_ctl07_divDropDown_ctl(0[2-9]|[1-9][0-9]*)')})]

        ret = pd.DataFrame()
        for route in routes:

            ctrl_dict['ctl00$MainContent$ssrsReportViewer$ctl08$ctl07$divDropDown$ctl01$HiddenIndices'] = \
                ','.join([str(x) for x in range(len(routes))])
            ctrl_dict['ctl00$MainContent$ssrsReportViewer$ctl08$ctl07$txtValue'] = route
            ctrl_dict['__EVENTTARGET'] = 'ctl00$MainContent$ssrsReportViewer$ctl08$ctl07'
            ctrl_dict['ctl00$MainContent$scriptManager'] = \
                'ctl00$MainContent$scriptManager|ctl00$MainContent$ssrsReportViewer$ctl08$ctl07'

            resp = self._make_response_and_submit(ctrl_dict, html)

            # turn the values in the page into a dictionary
            resp_dict = self.parse_ltiv_data(resp.decode())

            soup = BeautifulSoup(resp, features='html.parser')

            stops = [x.text.replace('\xa0', ' ') for x in soup.find_all('label', {
                'for':
                    re.compile(r'ctl00_MainContent_ssrsReportViewer_ctl08_ctl09_divDropDown_ctl(0[2-9]|[1-9][0-9]*)')})]

            # Setup the required values
            ctrl_dict['ctl00$MainContent$ssrsReportViewer$ctl08$ctl09$txtValue'] = ','.join(stops)
            ctrl_dict['ctl00$MainContent$ssrsReportViewer$ctl08$ctl09$divDropDown$ctl01$HiddenIndices'] = \
                ','.join([str(x) for x in range(len(stops))])
            ctrl_dict['__VIEWSTATE'] = resp_dict['__VIEWSTATE'][0]
            ctrl_dict['__EVENTVALIDATION'] = resp_dict['__EVENTVALIDATION'][0]
            ctrl_dict['__EVENTTARGET'] = ''
            ctrl_dict['ctl00$MainContent$ssrsReportViewer$ctl08$ctl00'] = 'View Report'
            ctrl_dict['ctl00$MainContent$ssrsReportViewer$ctl14'] = 'ltr'
            ctrl_dict['ctl00$MainContent$scriptManager'] = \
                'ctl00$MainContent$scriptManager|ctl00$MainContent$ssrsReportViewer$ctl08$ctl00'

            resp = self._make_response_and_submit(ctrl_dict, html)
            resp_dict = self.parse_ltiv_data(resp.decode())

            # Setup the required values
            ctrl_dict['null'] = '100'
            ctrl_dict['ctl00$MainContent$ssrsReportViewer$ctl09$ctl03$ctl00'] = ''
            ctrl_dict['__EVENTTARGET'] = 'ctl00$MainContent$ssrsReportViewer$ctl13$Reserved_AsyncLoadTarget'
            ctrl_dict['__VIEWSTATE'] = resp_dict['__VIEWSTATE'][0]
            ctrl_dict['__EVENTVALIDATION'] = resp_dict['__EVENTVALIDATION'][0]
            ctrl_dict['__EVENTTARGET'] = 'ctl00$MainContent$ssrsReportViewer$ctl13$Reserved_AsyncLoadTarget'
            ctrl_dict['ctl00$MainContent$scriptManager'] = \
                'ctl00$MainContent$scriptManager|ctl00$MainContent$ssrsReportViewer$ctl13$Reserved_AsyncLoadTarget'
            ctrl_dict['ctl00$MainContent$ssrsReportViewer$ctl09$ctl00$CurrentPage'] = ''

            self._make_response_and_submit(ctrl_dict, html)
            csv_data = self._download_csv(resp)

            # Ridesystems puts three datasets in this single CSV, so we need to find the end of the first dataset
            index = csv_data.text.find('\r\n\r\nDate,')
            ret_tmp = pd.read_csv(StringIO(csv_data.text[:index]), delimiter=',', skiprows=[0, 1, 2, 3, 4, 5, 6],
                                  names=['date', 'route', 'stop', 'blockid', 'scheduledarrivaltime',
                                         'actualarrivaltime',
                                         'scheduleddeparturetime', 'actualdeparturetime', 'ontimestatus', 'vehicle'],
                                  parse_dates=['date'], dtype=str, )

            ret_tmp['scheduledarrivaltime'] = pd.to_datetime(ret_tmp['scheduledarrivaltime'],
                                                             format='%I:%M:%S %p').dt.time
            ret_tmp['actualarrivaltime'] = pd.to_datetime(ret_tmp['actualarrivaltime'],
                                                          format='%I:%M:%S %p').dt.time
            ret_tmp['scheduleddeparturetime'] = pd.to_datetime(ret_tmp['scheduleddeparturetime'],
                                                               format='%I:%M:%S %p').dt.time
            ret_tmp['actualdeparturetime'] = pd.to_datetime(ret_tmp['actualdeparturetime'],
                                                            format='%I:%M:%S %p').dt.time
            ret_tmp['vehicle'] = ret_tmp['vehicle'].astype(str)
            ret_tmp['route'] = ret_tmp['route'].str.split(" ", n=1).str[0]

            if ret.empty:
                ret = ret_tmp
            else:
                pd.concat([ret, ret_tmp])

        return ret

    def get_runtimes(self, start_date: date, end_date: date) -> pd.DataFrame:
        """
        Pulls the on time performance data
        :param start_date: The start date to search, inclusive. Searches starting from 12:00 AM
        :param end_date: The end date to search, inclusive. Searches ending at 11:59:59 PM
        :return: Returns a dataframe with 'route', 'vehicle', 'start_time', and 'end_time'
        """
        logger.info("Getting runtime report for {} to {}", start_date, end_date)
        # Pull the page the first time to get the form that we will need to resubmit a few times
        soup, html = self._select_form('/Secure/Admin/Reports/ReportViewer.aspx?Path=%2f'
                                       'OldRidesystems%2fGeneral+Reports%2fVehicle_Assignment_Report_Ver2')

        ctrl_dict: Dict[str, Union[str, List]] = {
            'ctl00$MainContent$scriptManager':
                'ctl00$MainContent$scriptManager|ctl00$MainContent$ssrsReportViewer$ctl08$ctl00',
            '__VIEWSTATE': soup.find('input', {'name': '__VIEWSTATE'})['value'],
            '__VIEWSTATEGENERATOR': soup.find('input', {'name': '__VIEWSTATEGENERATOR'})['value'],
            '__EVENTVALIDATION': soup.find('input', {'name': '__EVENTVALIDATION'})['value'],
            'ctl00$MainContent$ssrsReportViewer$ctl15': 'standards',
            'ctl00$MainContent$ssrsReportViewer$AsyncWait$HiddenCancelField': 'False',
            # Start Date
            'ctl00$MainContent$ssrsReportViewer$ctl08$ctl03$txtValue': start_date.strftime('%#m/%#d/%Y'),
            # End Date
            'ctl00$MainContent$ssrsReportViewer$ctl08$ctl05$txtValue': end_date.strftime('%#m/%#d/%Y 11:59:59 PM'),
            # Group By
            'ctl00$MainContent$ssrsReportViewer$ctl08$ctl07$ddValue': ['1'],
            # Routes
            'ctl00$MainContent$ssrsReportViewer$ctl08$ctl09$txtValue':
                ",".join([i.text.replace('\xa0', ' ') for i in soup.find_all('label', {
                    'for': re.compile(r'ctl00_MainContent_ssrsReportViewer_ctl08_ctl09_divDropDown_ctl..')}) if
                          i.text.replace('\xa0', ' ') != '(Select All)']),
            'ctl00$MainContent$ssrsReportViewer$ctl08$ctl09$divDropDown$ctl01$HiddenIndices': '0,1,2,3,4',
            'ctl00$MainContent$ssrsReportViewer$ToggleParam$collapse': 'false',
            'ctl00$MainContent$ssrsReportViewer$ctl11$collapse': 'false',
            'ctl00$MainContent$ssrsReportViewer$ctl13$VisibilityState$ctl00': 'None',
            'ctl00$MainContent$ssrsReportViewer$ctl13$ReportControl$ctl04': '100',
            '__ASYNCPOST': 'true',
            'ctl00$MainContent$ssrsReportViewer$ctl08$ctl00': 'View Report'
        }

        resp = self._make_response_and_submit(ctrl_dict, html)
        csv_data = self._download_csv(resp)

        ret = pd.read_csv(StringIO(csv_data.text), skiprows=[0, 1, 2, 3], usecols=[1, 5, 6, 7],
                          names=['route', 'vehicle', 'start_time', 'end_time'], parse_dates=['start_time', 'end_time'])
        ret['route'] = ret['route'].str.split(" ", n=1).str[0]

        return ret

    def get_ridership(self, start_date: date, end_date: date) -> pd.DataFrame:
        """
        Pulls the raw ridership data report
        :param start_date: The start date to search, inclusive. Searches starting from 12:00 AM
        :param end_date: The end date to search, inclusive. Searches ending at 11:59:59 PM
        """
        logger.info('Getting raw ridership for dates {} and {}', start_date, end_date)
        soup, html = self._select_form(
            '/Secure/Admin/Reports/ReportViewer.aspx?Path=%2fOldRidesystems%2fRidership%2fRaw+Ridership')

        ctrl_dict: Dict[str, Union[str, List]] = {
            'ctl00$MainContent$scriptManager':
                'ctl00$MainContent$scriptManager|ctl00$MainContent$ssrsReportViewer$ctl13$Reserved_AsyncLoadTarget',
            '__EVENTTARGET': 'ctl00$MainContent$ssrsReportViewer$ctl13$Reserved_AsyncLoadTarget',
            '__VIEWSTATE': soup.find('input', {'name': '__VIEWSTATE'})['value'],
            '__VIEWSTATEGENERATOR': soup.find('input', {'name': '__VIEWSTATEGENERATOR'})['value'],
            '__EVENTVALIDATION': soup.find('input', {'name': '__EVENTVALIDATION'})['value'],
            'ctl00$MainContent$ssrsReportViewer$ctl15': 'standards',
            'ctl00$MainContent$ssrsReportViewer$AsyncWait$HiddenCancelField': 'False',
            'ctl00$MainContent$ssrsReportViewer$ctl08$ctl03$txtValue': start_date.strftime('%#m/%#d/%Y'),
            'ctl00$MainContent$ssrsReportViewer$ctl08$ctl05$txtValue': end_date.strftime('%#m/%#d/%Y 11:59:59 PM'),
            'ctl00$MainContent$ssrsReportViewer$ToggleParam$collapse': 'false',
            'null': '100',
            'ctl00$MainContent$ssrsReportViewer$ctl11$collapse': 'false',
            'ctl00$MainContent$ssrsReportViewer$ctl13$VisibilityState$ctl00': 'None',
            'ctl00$MainContent$ssrsReportViewer$ctl13$ReportControl$ctl04': '100',
            '__ASYNCPOST': 'true'
        }

        resp = self._make_response_and_submit(ctrl_dict, html)
        csv_data = self._download_csv(resp)

        dtypes = {'vehicle': str,
                  'route': str,
                  'stop': str,
                  'latitude': float,
                  'longitude': float,
                  'datetime': str,
                  'entries': int,
                  'exits': int
                  }
        ret = pd.read_csv(StringIO(csv_data.text), usecols=[1, 4, 6, 7, 8, 10, 11, 12], parse_dates=['datetime'],
                          skiprows=[0], keep_default_na=False, names=dtypes.keys())
        ret['route'] = ret['route'].str.split(" ", n=1).str[0]
        ret['datetime'] = ret['datetime'].dt.floor('Min')

        ret = ret.groupby(['vehicle', 'route', 'stop', 'datetime']).aggregate({
            'vehicle': 'first',
            'route': 'first',
            'stop': 'first',
            'latitude': 'first',
            'longitude': 'first',
            'datetime': 'first',
            'entries': 'sum',
            'exits': 'sum'
        })

        # remove rows not on a route, and return
        return ret[ret['route'] != '']

    @staticmethod
    def parse_ltiv_data(data: str) -> Dict[str, Tuple[str, str]]:
        """
        Parses the data that comes back from the aspx pages. Its in the format LENGTH|TYPE|ID|VALUE
        :param data:
        :return: Returns {ID: (VALUE, TYPE), ID: (VALUE, TYPE)}
        """

        def get_next_element(idata: str, ilength: int = None) -> Tuple[str, str]:
            """Parser that pulls off an element to the next delimiter, and optionally will read ilength bytes"""
            if ilength is not None:
                assert ilength < len(idata) and idata[ilength] == '|', \
                    "Malformed input. Expected delimiter where there wasn't one. idata: {}".format(idata[:100])
                iret = idata[:ilength]
                idata = idata[ilength + 1:]  # drop the delimiter
                return iret, idata
            return get_next_element(idata, idata.index('|'))

        ret = {}

        while data:
            s_length, data = get_next_element(data)
            length = int(s_length)
            data_type, data = get_next_element(data)
            data_id, data = get_next_element(data)
            value, data = get_next_element(data, length)

            ret[data_id] = (value, data_type)
        return ret

    def _download_csv(self, resp: bytes) -> requests.Response:
        """
        Gets the URL to the CSV file with the results of a search
        :param resp: The contents of a webpage that will contain the download url
        """
        response_url_base_group = re.search(b'"ExportUrlBase":"(.*?)"', resp)
        assert response_url_base_group is not None
        response_url_base = response_url_base_group.group(1).decode('utf-8').replace('\\u0026', '&')

        csv_data = requests.get("{}{}CSV".format(self.baseurl, response_url_base),
                                cookies=self.browser.cookiejar,
                                headers={
                                    'referer': '{}/Secure/Admin/Reports/ReportViewer.aspx?Path=%2fOldRidesystems%2f'
                                               'Ridership%2fAll+Ridership+By+Vehicle'.format(self.baseurl),
                                    'User-Agent': HEADERS},
                                )

        assert csv_data, "Request failed with status code {}".format(csv_data.status_code)
        logger.debug("Got %s bytes of data", len(csv_data.text))

        return csv_data

    def _select_form(self, url_path) -> Tuple[BeautifulSoup, str]:
        """
        Session setup, by opening the inital page, and selecting the proper form
        :param url_path: Suffix of the URL of the report to pull and select. Will be appended to self.baseurl
        :return: Tuple with the response as a BeautifulSoup object, and the html of the form we selected
        """
        resp = self.browser.open("{}{}".format(self.baseurl, url_path)).read()
        soup = BeautifulSoup(resp, features="html.parser")
        html = soup.find('form', id='aspnetForm').prettify().encode('utf8')

        self.browser.select_form('aspnetForm')
        self.browser.form.set_all_readonly(False)

        return soup, html

    def _set_controls(self, ctrl_dict: Dict[str, Any]) -> None:
        for ctrl_id, val in ctrl_dict.items():
            try:
                ctrl = self.browser.form.find_control(name=ctrl_id)
                ctrl.disabled = False
                ctrl.value = val
            except mechanize.ControlNotFoundError:
                self.browser.form.new_control('hidden', ctrl_id, {'value': val})
        self.browser.form.fixup()
        self._log_controls()

    def _log_controls(self) -> None:
        logger.debug('\n'.join(
            ['%s: %s *%s*' % (c.name, c.value, c.disabled) if c.disabled else '%s: %s' % (c.name, c.value) for c in
             self.browser.form.controls]))
